@@ -184,8 +184,13 @@ typedef struct pocl_hsa_device_data_s {
 
   /* if agent supports async handlers*/
   int have_wait_any;
+
+  /* compilation lock */
+  pocl_lock_t pocl_hsa_compilation_lock;
 } pocl_hsa_device_data_t;
 
+
+
 void
 pocl_hsa_init_device_ops(struct pocl_device_ops *ops)
 {
@@ -239,7 +244,7 @@ pocl_hsa_abort_on_hsa_error(hsa_status_t status,
   if (status != HSA_STATUS_SUCCESS)
     {
       hsa_status_string(status, &str);
-      POCL_MSG_PRINT2(func, line, "Error from HSA Runtime call:\n");
+      POCL_MSG_PRINT2(HSA, func, line, "Error from HSA Runtime call:\n");
       POCL_ABORT("%s", str);
     }
 }
@@ -258,7 +263,7 @@ pocl_hsa_abort_on_pthread_error(int status,
 {
   if (status != 0)
     {
-      POCL_MSG_PRINT2(func, line, "Error from pthread call:\n");
+      POCL_MSG_PRINT2(HSA, func, line, "Error from pthread call:\n");
       POCL_ABORT("%s", strerror(status));
     }
 }
@@ -269,8 +274,7 @@ pocl_hsa_abort_on_pthread_error(int status,
                                                             #code);
 
 static hsa_agent_t hsa_agents[MAX_HSA_AGENTS];
-static intptr_t last_assigned_agent = 0;
-static int found_hsa_agents = 0;
+static unsigned found_hsa_agents = 0;
 
 static hsa_status_t
 pocl_hsa_get_agents_callback(hsa_agent_t agent, void *data)
@@ -440,9 +444,9 @@ get_hsa_device_features(char* dev_name, struct _cl_device_id* dev)
 }
 
 void
-pocl_hsa_init_device_infos(struct _cl_device_id* dev)
+pocl_hsa_init_device_infos(unsigned j, struct _cl_device_id* dev)
 {
-  pocl_basic_init_device_infos (dev);
+  pocl_basic_init_device_infos (j, dev);
 
   SETUP_DEVICE_CL_VERSION(HSA_DEVICE_CL_VERSION_MAJOR,
                           HSA_DEVICE_CL_VERSION_MINOR)
@@ -454,10 +458,10 @@ pocl_hsa_init_device_infos(struct _cl_device_id* dev)
   dev->local_as_id = 3;
   dev->constant_as_id = 2;
 
-  assert(found_hsa_agents > 0);
-  assert(last_assigned_agent < found_hsa_agents);
-  dev->data = (void*)last_assigned_agent;
-  hsa_agent_t agent = hsa_agents[last_assigned_agent++];
+  assert (found_hsa_agents > 0);
+  assert (j < found_hsa_agents);
+  dev->data = (void*)(uintptr_t)j;
+  hsa_agent_t agent = hsa_agents[j];
 
   uint32_t cache_sizes[4];
   HSA_CHECK(hsa_agent_get_info (agent, HSA_AGENT_INFO_CACHE_SIZE,
@@ -583,9 +587,8 @@ pocl_hsa_probe(struct pocl_device_ops *ops)
   HSA_CHECK(hsa_iterate_agents(pocl_hsa_get_agents_callback, NULL));
 
   POCL_MSG_PRINT_INFO("pocl-hsa: found %d agents.\n", found_hsa_agents);
-  last_assigned_agent = 0;
 
-  return found_hsa_agents;
+  return (int)found_hsa_agents;
 }
 
 static void
@@ -596,8 +599,8 @@ hsa_queue_callback(hsa_status_t status, hsa_queue_t *q, void* data) {
 /* driver pthread prototype */
 void * pocl_hsa_driver_pthread (void *cldev);
 
-void
-pocl_hsa_init (cl_device_id device, const char* parameters)
+cl_int
+pocl_hsa_init (unsigned j, cl_device_id device, const char* parameters)
 {
   pocl_hsa_device_data_t *d;
 
@@ -605,6 +608,8 @@ pocl_hsa_init (cl_device_id device, const char* parameters)
 
   d = (pocl_hsa_device_data_t *) calloc (1, sizeof(pocl_hsa_device_data_t));
 
+  POCL_INIT_LOCK (d->pocl_hsa_compilation_lock);
+
   intptr_t agent_index = (intptr_t)device->data;
   d->agent.handle = hsa_agents[agent_index].handle;
   device->data = d;
@@ -691,6 +696,7 @@ pocl_hsa_init (cl_device_id device, const char* parameters)
   d->exit_driver_thread = 0;
   PTHREAD_CHECK(pthread_create(&d->driver_pthread_id, NULL,
                  &pocl_hsa_driver_pthread, device));
+  return CL_SUCCESS;
 }
 
 static void*
@@ -934,41 +940,6 @@ setup_kernel_args (pocl_hsa_device_data_t *d,
 #endif
 }
 
-/*
- * This replaces a simple system(), because system() was causing issues
- * (gpu lockups) when compiling code (via compile_parallel_bc_to_brig)
- * with OpenCL 2.0 atomics (like CalcPie from AMD SDK).
- * The reason of lockups is unknown (yet).
- */
-static int
-run_command(char* args[])
-{
-  POCL_MSG_PRINT_INFO("Launching: %s\n", args[0]);
-#ifdef HAVE_VFORK
-  pid_t p = vfork();
-#elif defined(HAVE_FORK)
-  pid_t p = fork();
-#else
-#error Must have fork() or vfork() system calls for HSA
-#endif
-  if (p == 0)
-    {
-      return execv(args[0], args);
-    }
-  else
-    {
-      if (p < 0)
-        return -1;
-      int status;
-      if (waitpid(p, &status, 0) < 0)
-        POCL_ABORT("pocl-hsa: waitpid() itself failed.\n");
-      if (WIFEXITED(status))
-        return WEXITSTATUS(status);
-      else
-        return -2;
-    }
-}
-
 static int
 compile_parallel_bc_to_brig(char* brigfile, cl_kernel kernel,
                             cl_device_id device) {
@@ -998,14 +969,14 @@ compile_parallel_bc_to_brig(char* brigfile, cl_kernel kernel,
 
       char* args1[] = { LLVM_LLC, "-O2", "-march=hsail64", "-filetype=asm",
                         "-o", hsailfile, parallel_bc_path, NULL };
-      if ((error = run_command(args1)))
+      if ((error = pocl_run_command (args1)))
         {
           POCL_MSG_PRINT_INFO("pocl-hsa: llc exit status %i\n", error);
           return error;
         }
 
       char* args2[] = { HSAIL_ASM, "-o", brigfile, hsailfile, NULL };
-      if ((error = run_command(args2)))
+      if ((error = pocl_run_command (args2)))
         {
           POCL_MSG_PRINT_INFO("pocl-hsa: HSAILasm exit status %i\n", error);
           return error;
@@ -1027,12 +998,26 @@ pocl_hsa_compile_kernel (_cl_command_node *cmd, cl_kernel kernel,
 
   hsa_executable_t final_obj;
 
+  POCL_LOCK (d->pocl_hsa_compilation_lock);
+
+  int error = pocl_llvm_generate_workgroup_function (device, kernel,
+                                        cmd->command.run.local_x,
+                                        cmd->command.run.local_y,
+                                        cmd->command.run.local_z);
+  if (error)
+    {
+      POCL_MSG_PRINT_GENERAL ("HSA: pocl_llvm_generate_workgroup_function()"
+                              " failed for kernel %s\n", kernel->name);
+      assert (error == 0);
+    }
+
   unsigned i;
   for (i = 0; i<HSA_KERNEL_CACHE_SIZE; i++)
     if (d->kernel_cache[i].kernel == kernel)
       {
         POCL_MSG_PRINT_INFO("kernel.hsa_exe found in"
                             " kernel cache, returning\n");
+        POCL_UNLOCK (d->pocl_hsa_compilation_lock);
         return;
       }
 
@@ -1136,6 +1121,7 @@ pocl_hsa_compile_kernel (_cl_command_node *cmd, cl_kernel kernel,
        kernel_symbol, HSA_EXECUTABLE_SYMBOL_INFO_KERNEL_KERNARG_SEGMENT_SIZE,
        &d->kernel_cache[i].args_segment_size));
 
+  POCL_UNLOCK (d->pocl_hsa_compilation_lock);
 }
 
 void
@@ -1165,6 +1151,8 @@ pocl_hsa_uninit (cl_device_id device)
 
   PTHREAD_CHECK(pthread_mutex_destroy(&d->list_mutex));
 
+  POCL_DESTROY_LOCK (d->pocl_hsa_compilation_lock);
+
   POCL_MEM_FREE(d);
   device->data = NULL;
 }
@@ -1205,12 +1193,11 @@ pocl_hsa_submit (_cl_command_node *node, cl_command_queue cq)
 
   POCL_LOCK_OBJ (node->event);
   PTHREAD_CHECK(pthread_mutex_lock(&d->list_mutex));
-  POCL_UPDATE_EVENT_SUBMITTED(&node->event);
 
-  /* this "ready" consept to ensure that command is pushed only once */
-  if (!(node->ready) && pocl_command_is_ready(node->event))
+  node->ready = 1;
+  if (pocl_command_is_ready (node->event))
     {
-      node->ready = 1;
+      POCL_UPDATE_EVENT_SUBMITTED (node->event);
       PN_ADD(d->ready_list, node->event);
       added_to_readylist = 1;
     }
@@ -1240,29 +1227,32 @@ pocl_hsa_join (cl_device_id device, cl_command_queue cq)
     }
   cl_event event = cq->last_event.event;
   assert(event);
-  POCL_LOCK_OBJ(event);
+  POCL_LOCK_OBJ (event);
+  ++event->pocl_refcount;
   POCL_UNLOCK_OBJ (cq);
 
   POCL_MSG_PRINT_INFO("pocl-hsa: device->join on event %u\n", event->id);
 
-  if (event->status == CL_COMPLETE)
+  if (event->status <= CL_COMPLETE)
     {
       POCL_MSG_PRINT_INFO("pocl-hsa: device->join: last event (%u) in queue"
                           " exists, but is complete\n", event->id);
-      POCL_UNLOCK_OBJ(event);
+      --event->pocl_refcount;
+      POCL_UNLOCK_OBJ (event);
       return;
     }
 
-  while (event->status != CL_COMPLETE)
+  while (event->status > CL_COMPLETE)
     {
       pocl_hsa_event_data_t *e_d = (pocl_hsa_event_data_t *)event->data;
       PTHREAD_CHECK (pthread_cond_wait (&e_d->event_cond, &event->pocl_lock));
     }
-  POCL_UNLOCK_OBJ(event);
-
   POCL_MSG_PRINT_INFO("pocl-hsa: device->join on event %u finished"
                       " with status: %i\n", event->id, event->status);
-  assert(event->status == CL_COMPLETE);
+
+  assert (event->status <= CL_COMPLETE);
+  --event->pocl_refcount;
+  POCL_UNLOCK_OBJ (event);
 }
 
 void
@@ -1273,20 +1263,27 @@ pocl_hsa_flush (cl_device_id device, cl_command_queue cq)
 }
 
 void
-pocl_hsa_notify (cl_device_id device, cl_event event)
+pocl_hsa_notify (cl_device_id device, cl_event event, cl_event finished)
 {
   pocl_hsa_device_data_t *d = device->data;
-  POCL_LOCK_OBJ (event);
   _cl_command_node *node = event->command;
   int added_to_readylist = 0;
   POCL_MSG_PRINT_INFO("pocl-hsa: notify on event %u \n", event->id);
 
-  /* this "ready" consept to ensure that command is pushed only once */
-  if (!(node->ready) && pocl_command_is_ready(node->event))
+  if (finished->status < CL_COMPLETE)
     {
-      node->ready = 1;
-      if (event->status == CL_SUBMITTED)
+      POCL_UPDATE_EVENT_FAILED (event);
+      return;
+    }
+
+  if (!node->ready)
+    return;
+
+  if (pocl_command_is_ready (event))
+    {
+      if (event->status == CL_QUEUED)
         {
+          POCL_UPDATE_EVENT_SUBMITTED (event);
           PTHREAD_CHECK(pthread_mutex_lock(&d->list_mutex));
 
           size_t i = 0;
@@ -1307,10 +1304,10 @@ pocl_hsa_notify (cl_device_id device, cl_event event)
           PTHREAD_CHECK(pthread_mutex_unlock(&d->list_mutex));
         }
       else
-        POCL_MSG_WARN("node->ready was 0 but event %u is"
-                      " not submitted!\n", event->id);
+        POCL_MSG_WARN ("node->ready was 1 but event %u is"
+                       " not queued: status %i!\n",
+                       event->id, event->status);
     }
-  POCL_UNLOCK_OBJ (event);
 
   if (added_to_readylist)
     hsa_signal_subtract_relaxed(d->nudge_driver_thread, 1);
@@ -1328,7 +1325,7 @@ pocl_hsa_wait_event(cl_device_id device, cl_event event)
 {
   POCL_MSG_PRINT_INFO("pocl-hsa: device->wait_event on event %u\n", event->id);
   POCL_LOCK_OBJ (event);
-  if (event->status == CL_COMPLETE)
+  if (event->status <= CL_COMPLETE)
     {
       POCL_MSG_PRINT_INFO("pocl-hsa: device->wain_event: last event"
                           " (%u) in queue exists, but is complete\n", 
@@ -1336,7 +1333,7 @@ pocl_hsa_wait_event(cl_device_id device, cl_event event)
       POCL_UNLOCK_OBJ(event);
       return;
     }
-  while (event->status != CL_COMPLETE)
+  while (event->status > CL_COMPLETE)
     {
       pocl_hsa_event_data_t *e_d = (pocl_hsa_event_data_t *)event->data;
       PTHREAD_CHECK(pthread_cond_wait(&(e_d->event_cond), &event->pocl_lock));
@@ -1344,7 +1341,7 @@ pocl_hsa_wait_event(cl_device_id device, cl_event event)
   POCL_UNLOCK_OBJ(event);
 
   POCL_MSG_PRINT_INFO("event wait finished with status: %i\n", event->status);
-  assert(event->status == CL_COMPLETE);
+  assert (event->status <= CL_COMPLETE);
 }
 
 /* DRIVER PTHREAD part */
@@ -1457,8 +1454,8 @@ pocl_hsa_launch(pocl_hsa_device_data_t *d, cl_event event)
         = kernel_packet->completion_signal.handle;
     }
 
+  POCL_UPDATE_EVENT_RUNNING (event);
   POCL_UNLOCK_OBJ (event);
-  POCL_UPDATE_EVENT_RUNNING(&event);
 }
 
 static void
@@ -1492,11 +1489,7 @@ pocl_hsa_ndrange_event_finished (pocl_hsa_device_data_t *d, size_t i)
   hsa_memory_free(event_data->actual_kernargs);
 
   POCL_UNLOCK_OBJ (event);
-  POCL_UPDATE_EVENT_COMPLETE(&event);
-
-  uint64_t ns = event->time_end - event->time_start;
-  pocl_debug_print_duration(__func__,__LINE__,
-                            "HSA NDrange Kernel (host clock)", ns);
+  POCL_UPDATE_EVENT_COMPLETE (event);
 
 }
 
@@ -1567,8 +1560,9 @@ pocl_hsa_driver_pthread (void * cldev)
   pocl_hsa_device_pthread_data_t* dd = &d->driver_data;
 
   /* timeout counter, resets with each new queued kernel to 1/8, then
-   * exponentially increases by 40% up to about 3/4 of d->timeout */
-  uint64_t kernel_timeout_ns = d->timeout >> 3;
+   * exponentially increases by 40% up to about 3/4 of d->timeout.
+   * disabled for now */
+  /* uint64_t kernel_timeout_ns = d->timeout >> 3; */
 
   dd->running_list_size = 0;
   dd->last_queue = 0;
@@ -1600,11 +1594,13 @@ pocl_hsa_driver_pthread (void * cldev)
 
   while (1)
     {
-      // reset the signal
-
+      /* reset the signal. Disabled for now; see below */
+#if 0
       if (pocl_hsa_run_ready_commands(d))
         kernel_timeout_ns = d->timeout >> 3;
-
+#else
+      pocl_hsa_run_ready_commands(d);
+#endif
       if (d->exit_driver_thread)
         goto EXIT_PTHREAD;
 
@@ -1705,16 +1701,32 @@ pocl_hsa_update_event (cl_device_id device, cl_event event, cl_int status)
       if (event->queue->properties & CL_QUEUE_PROFILING_ENABLE)
         event->time_end = device->ops->get_timer_value(device->data);
 
+      uint64_t ns = event->time_end - event->time_start;
+      pocl_debug_print_duration (__func__,__LINE__,
+                                 "HSA NDrange Kernel (host clock)", ns);
+
+
       POCL_LOCK_OBJ (event);
       event->status = CL_COMPLETE;
-
       pthread_cond_signal(&e_d->event_cond);
+      POCL_UNLOCK_OBJ (event);
 
       device->ops->broadcast (event);
-      POCL_UNLOCK_OBJ (event);
       break;
     default:
-      assert("Invalid event status\n");
+      POCL_MSG_PRINT_INFO ("HSA: EVENT FAILED, event %d\n", event->id);
+      pocl_mem_objs_cleanup (event);
+      pocl_update_command_queue (event);
+
+      if (event->queue->properties & CL_QUEUE_PROFILING_ENABLE)
+        event->time_end = device->ops->get_timer_value (device->data);
+
+      POCL_LOCK_OBJ (event);
+      event->status = CL_FAILED;
+      pthread_cond_signal (&e_d->event_cond);
+      POCL_UNLOCK_OBJ (event);
+
+      device->ops->broadcast (event);
       break;
     }
 }
