@@ -22,11 +22,12 @@
 */
 #include "tce_common.h"
 #include "pocl_util.h"
+#include "pocl_cache.h"
+#include "pocl_llvm.h"
 #include "utlist.h"
 #include "common.h"
 
 #include "config.h"
-#include "install-paths.h"
 #include "pocl_runtime_config.h"
 #include "pocl_hash.h"
 
@@ -53,13 +54,8 @@
 #include <GlobalScope.hh>
 #include <Environment.hh>
 
-#include "pocl_cache.h"
-
 using namespace TTAMachine;
 
-#define COMMAND_LENGTH 256
-#define WORKGROUP_STRING_LENGTH 1024
-
 #include <algorithm>
 
 #define ALIGNMENT (std::max(ALIGNOF_FLOAT16, ALIGNOF_DOUBLE16))
@@ -72,6 +68,7 @@ TCEDevice::TCEDevice(cl_device_id dev, const char* adfName) :
   ready_list(NULL), command_list(NULL) {
   parent->data = this;
   pthread_mutex_init (&cq_lock, NULL);
+  POCL_INIT_LOCK(tce_compile_lock);
   dev->address_bits = 32;
   dev->autolocals_to_args = 1;
 #if defined(WORDS_BIGENDIAN) && WORDS_BIGENDIAN == 1
@@ -224,9 +221,9 @@ TCEDevice::tceccCommandLine
     }
   else 
     {
-      deviceMainSrc = TCEString(PKGDATADIR) + "/" + mainC;
+      deviceMainSrc = TCEString(POCL_INSTALL_PRIVATE_DATADIR) + "/" + mainC;
       assert(access(deviceMainSrc.c_str(), R_OK) == 0);
-      poclIncludePathSwitch = " -I " PKGDATADIR "/include";
+      poclIncludePathSwitch = " -I " POCL_INSTALL_PRIVATE_DATADIR "/include";
     }
 
   TCEString extraFlags = extraParams;
@@ -404,14 +401,29 @@ void
 pocl_tce_compile_kernel(_cl_command_node *cmd,
                         cl_kernel kernel, cl_device_id device)
 {
-
   if (cmd->type != CL_COMMAND_NDRANGE_KERNEL)
     return;
 
   void* data = cmd->device->data;
   TCEDevice *d = (TCEDevice*)data;
 
-  int error;
+  if (!kernel)
+    kernel = cmd->command.run.kernel;
+  if (!device)
+    device = cmd->device;
+
+  POCL_LOCK(d->tce_compile_lock);
+  int error = pocl_llvm_generate_workgroup_function(device, kernel,
+      cmd->command.run.local_x, cmd->command.run.local_y,
+      cmd->command.run.local_z);
+
+  if (error) {
+    POCL_UNLOCK(d->tce_compile_lock);
+    POCL_MSG_PRINT_GENERAL("TCE: pocl_llvm_generate_workgroup_function()"
+                           " failed for kernel %s\n", kernel->name);
+    assert(error == 0);
+  }
+
   char bytecode[POCL_FILENAME_LENGTH];
 
   assert(d != NULL);
@@ -437,6 +449,8 @@ pocl_tce_compile_kernel(_cl_command_node *cmd,
         POCL_ABORT("Error while running tcecc.");
       }
   }
+
+  POCL_UNLOCK(d->tce_compile_lock);
 }
 
 void
@@ -852,13 +866,13 @@ static void tce_command_scheduler (TCEDevice *d)
   while ((node = d->ready_list))
     {
       assert (pocl_command_is_ready(node->event));
-      CDL_DELETE (d->ready_list, node); 
-      pthread_mutex_unlock (&d->cq_lock);
+      CDL_DELETE (d->ready_list, node);
+      POCL_UNLOCK(d->cq_lock);
       assert (node->event->status == CL_SUBMITTED);
       if (node->type == CL_COMMAND_NDRANGE_KERNEL)
         pocl_tce_compile_kernel(node, NULL, NULL);
       pocl_exec_command(node);
-      pthread_mutex_lock (&d->cq_lock);
+      POCL_LOCK(d->cq_lock);
     }
     
   return;
@@ -868,15 +882,15 @@ void
 pocl_tce_submit (_cl_command_node *node, cl_command_queue /*cq*/)
 {
   TCEDevice *d = (TCEDevice*)node->device->data;
-  cl_event *event = &(node->event);
 
-  POCL_LOCK (d->cq_lock);
-  POCL_UPDATE_EVENT_SUBMITTED(event);
+  POCL_LOCK_OBJ(node->event);
+  node->ready = 1;
+  POCL_LOCK(d->cq_lock);
   pocl_command_push(node, &d->ready_list, &d->command_list);
+  POCL_UNLOCK_OBJ(node->event);
 
   tce_command_scheduler (d);
-
-  POCL_UNLOCK (d->cq_lock);
+  POCL_UNLOCK(d->cq_lock);
 
   return;
 }
@@ -890,14 +904,6 @@ void pocl_tce_flush (cl_device_id device, cl_command_queue /*cq*/)
   POCL_UNLOCK (d->cq_lock);
 }
 
-void
-pocl_tce_push_command (_cl_command_node *node)
-{
-  TCEDevice *d = (TCEDevice*)node->device->data;
-
-  pocl_command_push(node, &d->ready_list, &d->command_list);
-
-}
 
 void
 pocl_tce_join(cl_device_id device, cl_command_queue /*cq*/)
@@ -912,27 +918,30 @@ pocl_tce_join(cl_device_id device, cl_command_queue /*cq*/)
 }
 
 void
-pocl_tce_notify (cl_device_id device, cl_event event)
+pocl_tce_notify (cl_device_id device, cl_event event, cl_event finished)
 {
   TCEDevice *d = (TCEDevice*)device->data;
   _cl_command_node * volatile node = event->command;
-  
-  POCL_LOCK_OBJ (event);
-  if (!(node->ready) && pocl_command_is_ready(node->event))
-    {
-      node->ready = 1;
-      POCL_UNLOCK_OBJ (event);
-      if (node->event->status == CL_SUBMITTED)
-        {
-          POCL_LOCK (d->cq_lock);
-          CDL_DELETE (d->command_list, node);
-          CDL_PREPEND (d->ready_list, node);
-          tce_command_scheduler (d);
-          POCL_UNLOCK (d->cq_lock);
-        }
-      return;
+
+  if (finished->status < CL_COMPLETE) {
+    POCL_UPDATE_EVENT_FAILED(event);
+    return;
+  }
+
+  if (!node->ready)
+    return;
+
+  if (pocl_command_is_ready(event)) {
+    if (event->status == CL_QUEUED) {
+      POCL_UPDATE_EVENT_SUBMITTED(event);
+      POCL_LOCK(d->cq_lock);
+      CDL_DELETE(d->command_list, node);
+      CDL_PREPEND(d->ready_list, node);
+      tce_command_scheduler(d);
+      POCL_UNLOCK(d->cq_lock);
+    }
+    return;
     }
-  POCL_UNLOCK_OBJ (event);
 }
 
 void
